# Copyright 2024 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Data models for the slurmdbd daemon."""

from types import MappingProxyType

from ._model import (
    BaseModel,
    ColonSeparatorCallback,
    CommaSeparatorCallback,
    SlurmDictCallback,
    base_descriptors,
)


class SlurmdbdConfig(BaseModel):
    """Object representing the slurmdbd.conf configuration file.

    Top-level configuration definition and data validators sourced from
    the slurmdbd.conf manpage. `man slurmdbd.conf.5`
    """

    _primary_key = None
    _callbacks = MappingProxyType(
        {
            "auth_alt_types": CommaSeparatorCallback,
            "auth_alt_parameters": SlurmDictCallback,
            "communication_parameters": SlurmDictCallback,
            "debug_flags": CommaSeparatorCallback,
            "parameters": CommaSeparatorCallback,
            "plugin_dir": ColonSeparatorCallback,
            "private_data": CommaSeparatorCallback,
            "storage_parameters": SlurmDictCallback,
        }
    )

    archive_dir = property(*base_descriptors("ArchiveDir"))
    archive_events = property(*base_descriptors("ArchiveEvents"))
    archive_jobs = property(*base_descriptors("ArchiveJobs"))
    archive_resvs = property(*base_descriptors("ArchiveResvs"))
    archive_script = property(*base_descriptors("ArchiveScript"))
    archive_steps = property(*base_descriptors("ArchiveSteps"))
    archive_suspend = property(*base_descriptors("ArchiveSuspend"))
    archive_txn = property(*base_descriptors("ArchiveTXN"))
    archive_usage = property(*base_descriptors("ArchiveUsage"))
    auth_info = property(*base_descriptors("AuthInfo"))
    auth_alt_types = property(*base_descriptors("AuthAltTypes"))
    auth_alt_parameters = property(*base_descriptors("AuthAltParameters"))
    auth_type = property(*base_descriptors("AuthType"))
    commit_delay = property(*base_descriptors("CommitDelay"))
    communication_parameters = property(*base_descriptors("CommunicationParameters"))
    dbd_backup_host = property(*base_descriptors("DbdBackupHost"))
    dbd_addr = property(*base_descriptors("DbdAddr"))
    dbd_host = property(*base_descriptors("DbdHost"))
    dbd_port = property(*base_descriptors("DbdPort"))
    debug_flags = property(*base_descriptors("DebugFlags"))
    debug_level = property(*base_descriptors("DebugLevel"))
    debug_level_syslog = property(*base_descriptors("DebugLevelSyslog"))
    default_qos = property(*base_descriptors("DefaultQOS"))
    log_file = property(*base_descriptors("LogFile"))
    log_time_format = property(*base_descriptors("LogTimeFormat"))
    max_query_time_range = property(*base_descriptors("MaxQueryTimeRange"))
    message_timeout = property(*base_descriptors("MessageTimeout"))
    parameters = property(*base_descriptors("Parameters"))
    pid_file = property(*base_descriptors("PidFile"))
    plugin_dir = property(*base_descriptors("PluginDir"))
    private_data = property(*base_descriptors("PrivateData"))
    purge_event_after = property(*base_descriptors("PurgeEventAfter"))
    purge_job_after = property(*base_descriptors("PurgeJobAfter"))
    purge_resv_after = property(*base_descriptors("PurgeResvAfter"))
    purge_step_after = property(*base_descriptors("PurgeStepAfter"))
    purge_suspend_after = property(*base_descriptors("PurgeSuspendAfter"))
    purge_txn_after = property(*base_descriptors("PurgeTXNAfter"))
    purge_usage_after = property(*base_descriptors("PurgeUsageAfter"))
    slurm_user = property(*base_descriptors("SlurmUser"))
    storage_host = property(*base_descriptors("StorageHost"))
    storage_backup_host = property(*base_descriptors("StorageBackupHost"))
    storage_loc = property(*base_descriptors("StorageLoc"))
    storage_parameters = property(*base_descriptors("StorageParameters"))
    storage_pass = property(*base_descriptors("StoragePass"))
    storage_port = property(*base_descriptors("StoragePort"))
    storage_type = property(*base_descriptors("StorageType"))
    storage_user = property(*base_descriptors("StorageUser"))
    tcp_timeout = property(*base_descriptors("TCPTimeout"))
    track_slurmctld_down = property(*base_descriptors("TrackSlurmctldDown"))
    track_wc_key = property(*base_descriptors("TrackWCKey"))



# Copyright 2024 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Generic data models for the Slurm workload manager."""

import functools
from collections import UserList
from collections.abc import MutableMapping
from types import MappingProxyType
from typing import Any, Dict

from ._model import (
    BaseModel,
    ColonSeparatorCallback,
    CommaSeparatorCallback,
    ReasonCallback,
    SlurmDictCallback,
    assert_type,
    base_descriptors,
    nested_descriptors,
    primary_key_descriptors,
)

_node_descriptors = functools.partial(nested_descriptors, knob_key_alias="NodeName")
_frontend_descriptors = functools.partial(nested_descriptors, knob_key_alias="FrontendName")
_nodeset_descriptors = functools.partial(nested_descriptors, knob_key_alias="NodeSet")
_partition_descriptors = functools.partial(nested_descriptors, knob_key_alias="PartitionName")


class Node(BaseModel):
    """Object representing Node(s) definition in slurm.conf.

    Node definition and data validators sourced from
    the slurm.conf manpage. `man slurm.conf.5`
    """

    def __init__(self, **kwargs):
        super().__init__()
        self._register.update({kwargs.pop("NodeName"): {**kwargs}})

    _primary_key = "NodeName"
    _callbacks = MappingProxyType(
        {
            "cpu_spec_list": CommaSeparatorCallback,
            "features": CommaSeparatorCallback,
            "gres": CommaSeparatorCallback,
            "reason": ReasonCallback,
        }
    )

    node_name = property(*primary_key_descriptors())
    node_hostname = property(*_node_descriptors("NodeHostname"))
    node_addr = property(*_node_descriptors("NodeAddr"))
    bcast_addr = property(*_node_descriptors("BcastAddr"))
    boards = property(*_node_descriptors("Boards"))
    core_spec_count = property(*_node_descriptors("CoreSpecCount"))
    cores_per_socket = property(*_node_descriptors("CoresPerSocket"))
    cpu_bind = property(*_node_descriptors("CpuBind"))
    cpus = property(*_node_descriptors("CPUs"))
    cpu_spec_list = property(*_node_descriptors("CpuSpecList"))
    features = property(*_node_descriptors("Features"))
    gres = property(*_node_descriptors("Gres"))
    mem_spec_limit = property(*_node_descriptors("MemSpecLimit"))
    port = property(*_node_descriptors("Port"))
    procs = property(*_node_descriptors("Procs"))
    real_memory = property(*_node_descriptors("RealMemory"))
    reason = property(*_node_descriptors("Reason"))
    sockets = property(*_node_descriptors("Sockets"))
    sockets_per_board = property(*_node_descriptors("SocketsPerBoard"))
    state = property(*_node_descriptors("State"))
    threads_per_core = property(*_node_descriptors("ThreadsPerCore"))
    tmp_disk = property(*_node_descriptors("TmpDisk"))
    weight = property(*_node_descriptors("Weight"))


class DownNodes(BaseModel):
    """Object representing DownNodes definition in slurm.conf.

    DownNodes definition and data validators sourced from
    the slurm.conf manpage. `man slurm.conf.5`
    """

    _primary_key = None
    _callbacks = MappingProxyType(
        {
            "down_nodes": CommaSeparatorCallback,
            "reason": ReasonCallback,
        }
    )

    down_nodes = property(*base_descriptors("DownNodes"))
    reason = property(*base_descriptors("Reason"))
    state = property(*base_descriptors("State"))


class FrontendNode(BaseModel):
    """FrontendNode data model.

    FrontendNode definition and data validators sourced from
    the slurm.conf manpage. `man slurm.conf.5`
    """

    def __init__(self, **kwargs):
        super().__init__()
        self._register.update({kwargs.pop("FrontendName"): {**kwargs}})

    _primary_key = "FrontendName"
    _callbacks = MappingProxyType(
        {
            "allow_groups": CommaSeparatorCallback,
            "allow_users": CommaSeparatorCallback,
            "deny_groups": CommaSeparatorCallback,
            "deny_users": CommaSeparatorCallback,
            "reason": ReasonCallback,
        }
    )

    frontend_name = property(*primary_key_descriptors())
    frontend_addr = property(*_frontend_descriptors("FrontendAddr"))
    allow_groups = property(*_frontend_descriptors("AllowGroups"))
    allow_users = property(*_frontend_descriptors("AllowUsers"))
    deny_groups = property(*_frontend_descriptors("DenyGroups"))
    deny_users = property(*_frontend_descriptors("DenyUsers"))
    port = property(*_frontend_descriptors("Port"))
    reason = property(*_frontend_descriptors("Reason"))
    state = property(*_frontend_descriptors("State"))


class NodeSet(BaseModel):
    """Object representing NodeSet definition in slurm.conf.

    NodeSet definition and data validators sourced from
    the slurm.conf manpage. `man slurm.conf.5`
    """

    def __init__(self, **kwargs):
        super().__init__()
        self._register.update({kwargs.pop("NodeSet"): {**kwargs}})

    _primary_key = "NodeSet"
    _callbacks = MappingProxyType({"nodes": CommaSeparatorCallback})

    node_set = property(*primary_key_descriptors())
    feature = property(*_nodeset_descriptors("Feature"))
    nodes = property(*_nodeset_descriptors("Nodes"))


class Partition(BaseModel):
    """Object representing Partition definition in slurm.conf.

    Partition definition and data validators sourced from
    the slurm.conf manpage. `man slurm.conf.5`
    """

    def __init__(self, **kwargs):
        super().__init__()
        self._register.update({kwargs.pop("PartitionName"): {**kwargs}})

    _primary_key = "PartitionName"
    _callbacks = MappingProxyType(
        {
            "alloc_nodes": CommaSeparatorCallback,
            "allow_accounts": CommaSeparatorCallback,
            "allow_groups": CommaSeparatorCallback,
            "allow_qos": CommaSeparatorCallback,
            "deny_accounts": CommaSeparatorCallback,
            "deny_qos": CommaSeparatorCallback,
            "nodes": CommaSeparatorCallback,
            "tres_billing_weights": SlurmDictCallback,
        }
    )

    partition_name = property(*primary_key_descriptors())
    alloc_nodes = property(*_partition_descriptors("AllocNodes"))
    allow_accounts = property(*_partition_descriptors("AllowAccounts"))
    allow_groups = property(*_partition_descriptors("AllowGroups"))
    allow_qos = property(*_partition_descriptors("AllowQos"))
    alternate = property(*_partition_descriptors("Alternate"))
    cpu_bind = property(*_partition_descriptors("CpuBind"))
    default = property(*_partition_descriptors("Default"))
    default_time = property(*_partition_descriptors("DefaultTime"))
    def_cpu_per_gpu = property(*_partition_descriptors("DefCpuPerGPU"))
    def_mem_per_cpu = property(*_partition_descriptors("DefMemPerCPU"))
    def_mem_per_gpu = property(*_partition_descriptors("DefMemPerGPU"))
    def_mem_per_node = property(*_partition_descriptors("DefMemPerNode"))
    deny_accounts = property(*_partition_descriptors("DenyAccounts"))
    deny_qos = property(*_partition_descriptors("DenyQos"))
    disable_root_jobs = property(*_partition_descriptors("DisableRootJobs"))
    exclusive_user = property(*_partition_descriptors("ExclusiveUser"))
    grace_time = property(*_partition_descriptors("GraceTime"))
    hidden = property(*_partition_descriptors("Hidden"))
    lln = property(*_partition_descriptors("LLN"))
    max_cpus_per_node = property(*_partition_descriptors("MaxCPUsPerNode"))
    max_cpus_per_socket = property(*_partition_descriptors("MaxCPUsPerSocket"))
    max_mem_per_cpu = property(*_partition_descriptors("MaxMemPerCPU"))
    max_mem_per_node = property(*_partition_descriptors("MaxMemPerNode"))
    max_nodes = property(*_partition_descriptors("MaxNodes"))
    max_time = property(*_partition_descriptors("MaxTime"))
    min_nodes = property(*_partition_descriptors("MinNodes"))
    nodes = property(*_partition_descriptors("Nodes"))
    over_subscribe = property(*_partition_descriptors("OverSubscribe"))
    over_time_limit = property(*_partition_descriptors("OverTimeLimit"))
    power_down_on_idle = property(*_partition_descriptors("PowerDownOnIdle"))
    preempt_mode = property(*_partition_descriptors("PreemptMode"))
    priority_job_factor = property(*_partition_descriptors("PriorityJobFactor"))
    priority_tier = property(*_partition_descriptors("PriorityTier"))
    qos = property(*_partition_descriptors("QOS"))
    req_resv = property(*_partition_descriptors("ReqResv"))
    resume_timeout = property(*_partition_descriptors("ResumeTimeout"))
    root_only = property(*_partition_descriptors("RootOnly"))
    select_type_parameters = property(*_partition_descriptors("SelectTypeParameters"))
    state = property(*_partition_descriptors("State"))
    suspend_time = property(*_partition_descriptors("SuspendTime"))
    suspend_timeout = property(*_partition_descriptors("SuspendTimeout"))
    tres_billing_weights = property(*_partition_descriptors("TRESBillingWeights"))


class NodeMap(MutableMapping):
    """Map of Node names to dictionaries for composing `Node` objects."""

    def __init__(self, data: Dict[str, Dict[str, Any]]):
        self._register = data

    @assert_type(value=Node)
    def __setitem__(self, key: str, value: Node) -> None:
        if key != value.node_name:
            raise ValueError(f"{key} and {value.node_name} are not equal.")
        else:
            self._register.update(value.dict())

    def __delitem__(self, key: str) -> None:
        del self._register[key]

    def __getitem__(self, key: str) -> Node:
        try:
            node = self._register.get(key)
            return Node(NodeName=key, **node)
        except KeyError:
            raise KeyError(f"Node {key} is not defined.")

    def __len__(self):
        return len(self._register)

    def __iter__(self):
        return iter([Node(NodeName=k, **self._register[k]) for k in self._register.keys()])


class FrontendNodeMap(MutableMapping):
    """Map of FrontendNode names to dictionaries for composing `FrontendNode` objects."""

    def __init__(self, data: Dict[str, Dict[str, Any]]):
        self._register = data

    @assert_type(value=FrontendNode)
    def __setitem__(self, key: str, value: FrontendNode) -> None:
        if key != value.frontend_name:
            raise ValueError(f"{key} and {value.frontend_name} are not equal.")
        else:
            self._register.update(value.dict())

    def __delitem__(self, key: str) -> None:
        del self._register[key]

    def __getitem__(self, key: str) -> FrontendNode:
        try:
            frontend_node = self._register.get(key)
            return FrontendNode(FrontendName=key, **frontend_node)
        except KeyError:
            raise KeyError(f"FrontendNode {key} is not defined.")

    def __len__(self):
        return len(self._register)

    def __iter__(self):
        return iter(
            [FrontendNode(FrontendName=k, **self._register[k]) for k in self._register.keys()]
        )


class DownNodesList(UserList):
    """List of dictionaries for composing `DownNodes` objects."""

    def __getitem__(self, i):
        if isinstance(i, slice):
            return self.__class__(self.data[i])
        else:
            return DownNodes(**self.data[i])

    @assert_type(value=DownNodes)
    def __setitem__(self, i: int, value: DownNodes):
        super().__setitem__(i, value.dict())

    @assert_type(value=DownNodes)
    def __contains__(self, value):
        return value.dict() in self.data

    def __iter__(self):
        return iter([DownNodes(**data) for data in self.data])

    def __add__(self, other):
        if isinstance(other, DownNodesList):
            return self.__class__(self.data + other.data)
        elif isinstance(other, type(self.data)):
            # Cannot use `assert_type` here because isinstance does
            # not support using subscripted generics for runtime validation.
            for down_node in other:
                if not isinstance(down_node, DownNodes):
                    raise TypeError(f"{down_node} is not {type(DownNodes)}.")

            return self.__class__(self.data + other)

        return self.__class__(self.data + list(other))

    def __radd__(self, other):
        if isinstance(other, DownNodesList):
            return self.__class__(other.data + self.data)
        elif isinstance(other, type(self.data)):
            for down_node in other:
                if not isinstance(down_node, DownNodes):
                    raise TypeError(f"{down_node} is not {type(DownNodes)}.")

            return self.__class__(other + self.data)

        return self.__class__(list(other) + self.data)

    def __iadd__(self, other):
        if isinstance(other, DownNodesList):
            self.data += other.data
        elif isinstance(other, type(self.data)):
            for down_node in other:
                if not isinstance(down_node, DownNodes):
                    raise TypeError(f"{down_node} is not {type(DownNodes)}.")

            self.data += other
        else:
            if not isinstance(other, DownNodes):
                raise TypeError(f"{other} is not {type(DownNodes)}.")

            self.data += other
        return self

    @assert_type(value=DownNodes)
    def append(self, value: DownNodes):
        """Add DownNodes object to list of DownNodes."""
        super().append(value.dict())

    @assert_type(value=DownNodes)
    def insert(self, i, value):
        """Insert DownNodes object into list of DownNodes at the given index."""
        super().insert(i, value.dict())

    @assert_type(value=DownNodes)
    def remove(self, value):
        """Remove DownNodes object from list of DownNodes."""
        self.data.remove(value.dict())

    @assert_type(value=DownNodes)
    def count(self, value):
        """Count the numbers of occurrences for the given DownNodes object.

        Warnings:
            Each DownNodes object should only occur once (1). If the object
            occurs more than once, this can create BIG problems when trying to
            restart the Slurm daemons.
        """
        return self.data.count(value.dict())

    @assert_type(value=DownNodes)
    def index(self, value, *args):
        """Get the index of the give DownNodes object."""
        return self.data.index(value.dict(), *args)

    def extend(self, other):
        """Extend DownNodes list by appending DownNodes objects from the iterable."""
        if isinstance(other, DownNodesList):
            self.data.extend(other.data)
        else:
            for down_node in other:
                if not isinstance(down_node, DownNodes):
                    raise TypeError(f"{down_node} is not {type(DownNodes)}.")

            self.data.extend(other)


class NodeSetMap(MutableMapping):
    """Map of NodeSet names to dictionaries for composing `NodeSet` objects."""

    def __init__(self, data: Dict[str, Dict[str, Any]]):
        self._register = data

    @assert_type(value=NodeSet)
    def __setitem__(self, key: str, value: NodeSet) -> None:
        if key != value.node_set:
            raise ValueError(f"{key} and {value.node_set} are not equal.")
        else:
            self._register.update(value.dict())

    def __delitem__(self, key: str) -> None:
        del self._register[key]

    def __getitem__(self, key: str) -> NodeSet:
        try:
            node_set = self._register.get(key)
            return NodeSet(NodeSet=key, **node_set)
        except KeyError:
            raise KeyError(f"NodeSet {key} is not defined.")

    def __len__(self):
        return len(self._register)

    def __iter__(self):
        return iter([NodeSet(NodeSet=k, **self._register[k]) for k in self._register.keys()])


class PartitionMap(MutableMapping):
    """Map of Partition names to dictionaries for composing `Partition` objects."""

    def __init__(self, data: Dict[str, Dict[str, Any]]):
        self._register = data

    def __contains__(self, key: str):
        return key in self._register

    def __len__(self):
        return len(self._register)

    def __iter__(self):
        return iter(
            [Partition(PartitionName=k, **self._register[k]) for k in self._register.keys()]
        )

    def __getitem__(self, key: str) -> Partition:
        try:
            partition = self._register.get(key)
            return Partition(PartitionName=key, **partition)
        except KeyError:
            raise KeyError(f"Partition {key} is not defined.")

    @assert_type(value=Partition)
    def __setitem__(self, key: str, value: Partition) -> None:
        if key != value.partition_name:
            raise ValueError(f"{key} and {value.partition_name} are not equal.")
        else:
            self._register.update(value.dict())

    def __delitem__(self, key: str) -> None:
        del self._register[key]


class SlurmConfig(BaseModel):
    """Object representing the slurm.conf configuration file.

    Top-level configuration definition and data validators sourced from
    the slurm.conf manpage. `man slurm.conf.5`
    """

    _primary_key = None
    _callbacks = MappingProxyType(
        {
            "acct_storage_external_host": CommaSeparatorCallback,
            "acct_storage_param": SlurmDictCallback,
            "acct_storage_tres": CommaSeparatorCallback,
            "acct_store_flags": CommaSeparatorCallback,
            "auth_alt_types": CommaSeparatorCallback,
            "auth_alt_param": SlurmDictCallback,
            "auth_info": SlurmDictCallback,
            "bcast_exclude": CommaSeparatorCallback,
            "bcast_param": SlurmDictCallback,
            "cli_filter_plugins": CommaSeparatorCallback,
            "communication_params": SlurmDictCallback,
            "cpu_freq_def": CommaSeparatorCallback,
            "cpu_freq_governors": CommaSeparatorCallback,
            "debug_flags": CommaSeparatorCallback,
            "dependency_param": SlurmDictCallback,
            "federation_param": CommaSeparatorCallback,
            "health_check_node_state": CommaSeparatorCallback,
            "job_acct_gather_frequency": SlurmDictCallback,
            "job_comp_params": SlurmDictCallback,
            "job_submit_plugins": CommaSeparatorCallback,
            "launch_parameters": SlurmDictCallback,
            "licenses": CommaSeparatorCallback,
            "plugin_dir": ColonSeparatorCallback,
            "power_parameters": SlurmDictCallback,
            "preempt_mode": CommaSeparatorCallback,
            "preempt_param": SlurmDictCallback,
            "prep_plugins": CommaSeparatorCallback,
            "priority_weight_tres": SlurmDictCallback,
            "private_data": CommaSeparatorCallback,
            "prolog_flags": CommaSeparatorCallback,
            "propagate_resource_limits": CommaSeparatorCallback,
            "propagate_resource_limits_except": CommaSeparatorCallback,
            "scheduler_param": SlurmDictCallback,
            "scron_param": CommaSeparatorCallback,
            "slurmctld_param": SlurmDictCallback,
            "slurmd_param": CommaSeparatorCallback,
            "switch_param": SlurmDictCallback,
            "task_plugin": CommaSeparatorCallback,
            "task_plugin_param": SlurmDictCallback,
            "topology_param": CommaSeparatorCallback,
        }
    )

    include = property(*base_descriptors("Include"))
    accounting_storage_backup_host = property(*base_descriptors("AccountingStorageBackupHost"))
    accounting_storage_enforce = property(*base_descriptors("AccountingStorageEnforce"))
    account_storage_external_host = property(*base_descriptors("AccountStorageExternalHost"))
    accounting_storage_host = property(*base_descriptors("AccountingStorageHost"))
    accounting_storage_parameters = property(*base_descriptors("AccountingStorageParameters"))
    accounting_storage_pass = property(*base_descriptors("AccountingStoragePass"))
    accounting_storage_port = property(*base_descriptors("AccountingStoragePort"))
    accounting_storage_tres = property(*base_descriptors("AccountingStorageTRES"))
    accounting_storage_type = property(*base_descriptors("AccountingStorageType"))
    accounting_storage_user = property(*base_descriptors("AccountingStorageUser"))
    accounting_store_flags = property(*base_descriptors("AccountingStoreFlags"))
    acct_gather_node_freq = property(*base_descriptors("AcctGatherNodeFreq"))
    acct_gather_energy_type = property(*base_descriptors("AcctGatherEnergyType"))
    acct_gather_interconnect_type = property(*base_descriptors("AcctGatherInterconnectType"))
    acct_gather_filesystem_type = property(*base_descriptors("AcctGatherFilesystemType"))
    acct_gather_profile_type = property(*base_descriptors("AcctGatherProfileType"))
    allow_spec_resources_usage = property(*base_descriptors("AllowSpecResourcesUsage"))
    auth_alt_types = property(*base_descriptors("AuthAltTypes"))
    auth_alt_parameters = property(*base_descriptors("AuthAltParameters"))
    auth_info = property(*base_descriptors("AuthInfo"))
    auth_type = property(*base_descriptors("AuthType"))
    batch_start_timeout = property(*base_descriptors("BatchStartTimeout"))
    bcast_exclude = property(*base_descriptors("BcastExclude"))
    bcast_parameters = property(*base_descriptors("BcastParameters"))
    burst_buffer_type = property(*base_descriptors("BurstBufferType"))
    cli_filter_plugins = property(*base_descriptors("CliFilterPlugins"))
    cluster_name = property(*base_descriptors("ClusterName"))
    communication_parameters = property(*base_descriptors("CommunicationParameters"))
    complete_wait = property(*base_descriptors("CompleteWait"))
    core_spec_plugin = property(*base_descriptors("CoreSpecPlugin"))
    cpu_freq_def = property(*base_descriptors("CpuFreqDef"))
    cpu_freq_governors = property(*base_descriptors("CpuFreqGovernors"))
    cred_type = property(*base_descriptors("CredType"))
    debug_flags = property(*base_descriptors("DebugFlags"))
    def_cpu_per_gpu = property(*base_descriptors("DefCpuPerGPU"))
    def_mem_per_cpu = property(*base_descriptors("DefMemPerCPU"))
    def_mem_per_gpu = property(*base_descriptors("DefMemPerGPU"))
    def_mem_per_node = property(*base_descriptors("DefMemPerNode"))
    dependency_parameters = property(*base_descriptors("DependencyParameters"))
    disable_root_jobs = property(*base_descriptors("DisableRootJobs"))
    eio_timeout = property(*base_descriptors("EioTimeout"))
    enforce_part_limits = property(*base_descriptors("EnforcePartLimits"))
    epilog = property(*base_descriptors("Epilog"))
    epilog_msg_time = property(*base_descriptors("EpilogMsgTime"))
    epilog_slurmctld = property(*base_descriptors("EpilogSlurmctld"))
    ext_sensors_freq = property(*base_descriptors("ExtSensorsFreq"))
    ext_sensors_type = property(*base_descriptors("ExtSensorsType"))
    fair_share_dampening_factor = property(*base_descriptors("FairShareDampeningFactor"))
    federation_parameters = property(*base_descriptors("FederationParameters"))
    first_job_id = property(*base_descriptors("FirstJobId"))
    get_env_timeout = property(*base_descriptors("GetEnvTimeout"))
    gres_types = property(*base_descriptors("GresTypes"))
    group_update_force = property(*base_descriptors("GroupUpdateForce"))
    group_update_time = property(*base_descriptors("GroupUpdateTime"))
    gpu_freq_def = property(*base_descriptors("GpuFreqDef"))
    health_check_interval = property(*base_descriptors("HealthCheckInterval"))
    health_check_node_state = property(*base_descriptors("HealthCheckNodeState"))
    health_check_program = property(*base_descriptors("HealthCheckProgram"))
    inactive_limit = property(*base_descriptors("InactiveLimit"))
    interactive_step_options = property(*base_descriptors("InteractiveStepOptions"))
    job_acct_gather_type = property(*base_descriptors("JobAcctGatherType"))
    job_acct_gather_frequency = property(*base_descriptors("JobAcctGatherFrequency"))
    job_acct_gather_params = property(*base_descriptors("JobAcctGatherParams"))
    job_comp_host = property(*base_descriptors("JobCompHost"))
    job_comp_loc = property(*base_descriptors("JobCompLoc"))
    job_comp_params = property(*base_descriptors("JobCompParams"))
    job_comp_pass = property(*base_descriptors("JobCompPass"))
    job_comp_port = property(*base_descriptors("JobCompPort"))
    job_comp_type = property(*base_descriptors("JobCompType"))
    job_comp_user = property(*base_descriptors("JobCompUser"))
    job_container_type = property(*base_descriptors("JobContainerType"))
    job_file_append = property(*base_descriptors("JobFileAppend"))
    job_requeue = property(*base_descriptors("JobRequeue"))
    job_submit_plugins = property(*base_descriptors("JobSubmitPlugins"))
    kill_on_bad_exit = property(*base_descriptors("KillOnBadExit"))
    kill_wait = property(*base_descriptors("KillWait"))
    max_batch_requeue = property(*base_descriptors("MaxBatchRequeue"))
    node_features_plugins = property(*base_descriptors("NodeFeaturesPlugins"))
    launch_parameters = property(*base_descriptors("LaunchParameters"))
    licenses = property(*base_descriptors("Licenses"))
    log_time_format = property(*base_descriptors("LogTimeFormat"))
    mail_domain = property(*base_descriptors("MailDomain"))
    mail_prog = property(*base_descriptors("MailProg"))
    max_array_size = property(*base_descriptors("MaxArraySize"))
    max_job_count = property(*base_descriptors("MaxJobCount"))
    max_job_id = property(*base_descriptors("MaxJobId"))
    max_mem_per_cpu = property(*base_descriptors("MaxMemPerCPU"))
    max_mem_per_node = property(*base_descriptors("MaxMemPerNode"))
    max_node_count = property(*base_descriptors("MaxNodeCount"))
    max_step_count = property(*base_descriptors("MaxStepCount"))
    max_tasks_per_node = property(*base_descriptors("MaxTasksPerNode"))
    mcs_parameters = property(*base_descriptors("MCSParameters"))
    mcs_plugin = property(*base_descriptors("MCSPlugin"))
    message_timeout = property(*base_descriptors("MessageTimeout"))
    min_job_age = property(*base_descriptors("MinJobAge"))
    mpi_default = property(*base_descriptors("MpiDefault"))
    mpi_params = property(*base_descriptors("MpiParams"))
    over_time_limit = property(*base_descriptors("OverTimeLimit"))
    plugin_dir = property(*base_descriptors("PluginDir"))
    plug_stack_config = property(*base_descriptors("PlugStackConfig"))
    power_parameters = property(*base_descriptors("PowerParameters"))
    power_plugin = property(*base_descriptors("PowerPlugin"))
    preempt_mode = property(*base_descriptors("PreemptMode"))
    preempt_parameters = property(*base_descriptors("PreemptParameters"))
    preempt_type = property(*base_descriptors("PreemptType"))
    preempt_exempt_time = property(*base_descriptors("PreemptExemptTime"))
    prep_parameters = property(*base_descriptors("PrEpParameters"))
    prep_plugins = property(*base_descriptors("PrEpPlugins"))
    priority_calcp_period = property(*base_descriptors("PriorityCalcpPeriod"))
    priority_decay_half_life = property(*base_descriptors("PriorityDecayHalfLife"))
    priority_favor_small = property(*base_descriptors("PriorityFavorSmall"))
    priority_flags = property(*base_descriptors("PriorityFlags"))
    priority_max_age = property(*base_descriptors("PriorityMaxAge"))
    priority_parameters = property(*base_descriptors("PriorityParameters"))
    priority_site_factor_parameters = property(*base_descriptors("PrioritySiteFactorParameters"))
    priority_site_factor_plugin = property(*base_descriptors("PrioritySiteFactorPlugin"))
    priority_type = property(*base_descriptors("PriorityType"))
    priority_usage_reset_period = property(*base_descriptors("PriorityUsageResetPeriod"))
    priority_weight_age = property(*base_descriptors("PriorityWeightAge"))
    priority_weight_assoc = property(*base_descriptors("PriorityWeightAssoc"))
    priority_weight_fair_share = property(*base_descriptors("PriorityWeightFairShare"))
    priority_weight_job_size = property(*base_descriptors("PriorityWeightJobSize"))
    priority_weight_partition = property(*base_descriptors("PriorityWeightPartition"))
    priority_weight_qos = property(*base_descriptors("PriorityWeightQOS"))
    priority_weight_tres = property(*base_descriptors("PriorityWeightTRES"))
    private_data = property(*base_descriptors("PrivateData"))
    proctrack_type = property(*base_descriptors("ProctrackType"))
    prolog = property(*base_descriptors("Prolog"))
    prolog_epilog_timeout = property(*base_descriptors("PrologEpilogTimeout"))
    prolog_flags = property(*base_descriptors("PrologFlags"))
    prolog_slurmctld = property(*base_descriptors("PrologSlurmctld"))
    propagate_prio_process = property(*base_descriptors("PropagatePrioProcess"))
    propagate_resource_limits = property(*base_descriptors("PropagateResourceLimits"))
    propagate_resource_limits_except = property(*base_descriptors("PropagateResourceLimitsExcept"))
    reboot_program = property(*base_descriptors("RebootProgram"))
    reconfig_flags = property(*base_descriptors("ReconfigFlags"))
    requeue_exit = property(*base_descriptors("RequeueExit"))
    requeue_exit_hold = property(*base_descriptors("RequeueExitHold"))
    resume_fail_program = property(*base_descriptors("ResumeFailProgram"))
    resume_program = property(*base_descriptors("ResumeProgram"))
    resume_rate = property(*base_descriptors("ResumeRate"))
    resume_timeout = property(*base_descriptors("ResumeTimeout"))
    resv_epilog = property(*base_descriptors("ResvEpilog"))
    resv_over_run = property(*base_descriptors("ResvOverRun"))
    resv_prolog = property(*base_descriptors("ResvProlog"))
    return_to_service = property(*base_descriptors("ReturnToService"))
    route_plugin = property(*base_descriptors("RoutePlugin"))
    scheduler_parameters = property(*base_descriptors("SchedulerParameters"))
    scheduler_time_slice = property(*base_descriptors("SchedulerTimeSlice"))
    scheduler_type = property(*base_descriptors("SchedulerType"))
    scron_parameters = property(*base_descriptors("ScronParameters"))
    select_type = property(*base_descriptors("SelectType"))
    select_type_parameters = property(*base_descriptors("SelectTypeParameters"))
    slurmctld_addr = property(*base_descriptors("SlurmctldAddr"))
    slurmctld_debug = property(*base_descriptors("SlurmctldDebug"))
    slurmctld_host = property(*base_descriptors("SlurmctldHost"))
    slurmctld_log_file = property(*base_descriptors("SlurmctldLogFile"))
    slurmctld_parameters = property(*base_descriptors("SlurmctldParameters"))
    slurmctld_pid_file = property(*base_descriptors("SlurmctldPidFile"))
    slurmctld_port = property(*base_descriptors("SlurmctldPort"))
    slurmctld_primary_off_prog = property(*base_descriptors("SlurmctldPrimaryOffProg"))
    slurmctld_primary_on_prog = property(*base_descriptors("SlurmctldPrimaryOnProg"))
    slurmctld_syslog_debug = property(*base_descriptors("SlurmctldSyslogDebug"))
    slurmctld_timeout = property(*base_descriptors("SlurmctldTimeout"))
    slurmd_debug = property(*base_descriptors("SlurmdDebug"))
    slurmd_log_file = property(*base_descriptors("SlurmdLogFile"))
    slurmd_parameters = property(*base_descriptors("SlurmdParameters"))
    slurmd_pid_file = property(*base_descriptors("SlurmdPidFile"))
    slurmd_port = property(*base_descriptors("SlurmdPort"))
    slurmd_spool_dir = property(*base_descriptors("SlurmdSpoolDir"))
    slurmd_syslog_debug = property(*base_descriptors("SlurmdSyslogDebug"))
    slurmd_timeout = property(*base_descriptors("SlurmdTimeout"))
    slurmd_user = property(*base_descriptors("SlurmdUser"))
    slurm_sched_log_file = property(*base_descriptors("SlurmSchedLogFile"))
    slurm_sched_log_level = property(*base_descriptors("SlurmSchedLogLevel"))
    slurm_user = property(*base_descriptors("SlurmUser"))
    srun_epilog = property(*base_descriptors("SrunEpilog"))
    srun_port_range = property(*base_descriptors("SrunPortRange"))
    srun_prolog = property(*base_descriptors("SrunProlog"))
    state_save_location = property(*base_descriptors("StateSaveLocation"))
    suspend_exc_nodes = property(*base_descriptors("SuspendExcNodes"))
    suspend_exc_parts = property(*base_descriptors("SuspendExcParts"))
    suspend_exc_states = property(*base_descriptors("SuspendExcStates"))
    suspend_program = property(*base_descriptors("SuspendProgram"))
    suspend_rate = property(*base_descriptors("SuspendRate"))
    suspend_time = property(*base_descriptors("SuspendTime"))
    suspend_timeout = property(*base_descriptors("SuspendTimeout"))
    switch_parameters = property(*base_descriptors("SwitchParameters"))
    switch_type = property(*base_descriptors("SwitchType"))
    task_epilog = property(*base_descriptors("TaskEpilog"))
    task_plugin = property(*base_descriptors("TaskPlugin"))
    task_plugin_param = property(*base_descriptors("TaskPluginParam"))
    task_prolog = property(*base_descriptors("TaskProlog"))
    tcp_timeout = property(*base_descriptors("TCPTimeout"))
    tmp_fs = property(*base_descriptors("TmpFS"))
    topology_param = property(*base_descriptors("TopologyParam"))
    topology_plugin = property(*base_descriptors("TopologyPlugin"))
    track_wc_key = property(*base_descriptors("TrackWCKey"))
    tree_width = property(*base_descriptors("TreeWidth"))
    unkillable_step_program = property(*base_descriptors("UnkillableStepProgram"))
    unkillable_step_timeout = property(*base_descriptors("UnkillableStepTimeout"))
    use_pam = property(*base_descriptors("UsePAM"))
    vsize_factor = property(*base_descriptors("VSizeFactor"))
    wait_time = property(*base_descriptors("WaitTime"))
    x11_parameters = property(*base_descriptors("X11Parameters"))

    @property
    def nodes(self) -> NodeMap:
        """Get the current nodes in the Slurm configuration."""
        return NodeMap(self._register["nodes"])

    @property
    def frontend_nodes(self) -> FrontendNodeMap:
        """Get the current frontend nodes in the Slurm configuration."""
        return FrontendNodeMap(self._register["frontend_nodes"])

    @property
    def down_nodes(self) -> DownNodesList:
        """Get the current down nodes in the Slurm configuration."""
        return DownNodesList(self._register["down_nodes"])

    @property
    def node_sets(self) -> NodeSetMap:
        """Get the current node sets in the Slurm configuration."""
        return NodeSetMap(self._register["node_sets"])

    @property
    def partitions(self) -> PartitionMap:
        """Get the current partitions in the Slurm configuration."""
        return PartitionMap(self._register["partitions"])



# Copyright 2024 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Macros for Slurm workload manager data models."""

import copy
import functools
import inspect
import json
from abc import ABC, abstractmethod
from types import MappingProxyType
from typing import Any, Callable, Dict, NamedTuple, Optional


# Simple type checking decorator; used to verify input into Slurm data models
# without needing every method to contain an `if isinstance(...)` block.
def assert_type(*typed_args, **typed_kwargs):
    """Check the type of args and kwargs passed to a function/method."""

    def decorator(func: Callable):
        sig = inspect.signature(func)
        bound_types = sig.bind_partial(*typed_args, **typed_kwargs).arguments

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            bound_values = sig.bind(*args, **kwargs).arguments
            for name in bound_types.keys() & bound_values.keys():
                if not isinstance(bound_values[name], bound_types[name]):
                    raise TypeError(f"{bound_values[name]} is not {bound_types[name]}.")

            return func(*args, **kwargs)

        return wrapper

    return decorator


# Generate descriptors for Slurm configuration knobs.
# These descriptors are used for retrieving configuration values but
# also preserve the integrity of Slurm's loose pascal casing.
# The descriptors will use an internal _register dictionary to
# manage the parsed configuration knobs.
def base_descriptors(knob: str):
    """Generate descriptors for accessing configuration knob values.

    Args:
        knob: Configuration knob to generate descriptors for.
    """

    def getter(self):
        return self._register.get(knob, None)

    def setter(self, value):
        self._register[knob] = value

    def deleter(self):
        try:
            del self._register[knob]
        except KeyError:
            pass

    return getter, setter, deleter


# Nodes, FrontendNodes, DownNodes, NodeSets, and Partitions are represented
# as a Python dictionary with a primary key and nested dictionary when
# parsed in from the slurm.conf configuration file:
#
# {"node_1": {"NodeHostname": ..., "NodeAddr": ..., "CPUs", ...}}
#
# Since these models are parsed in this way, they need special descriptors
# for accessing the primary key (e.g. the NodeName), and sub values in the
# nested dictionary.
def primary_key_descriptors():
    """Generate descriptors for accessing a configuration knob key."""

    def getter(self):
        # There will only be a single key in _register,
        # so it's okay to return the first index. If the
        # primary key doesn't exist, return None.
        try:
            return list(self._register.keys())[0]
        except IndexError:
            return None

    def setter(self, value):
        old_primary = list(self._register.keys())[0]
        if old_primary:
            self._register[value] = self._register.pop(old_primary, {})
        else:
            self._register[value] = {}

    def deleter(self):
        try:
            primary_key = list(self._register.keys())[0]
            del self._register[primary_key]
        except IndexError:
            pass

    return getter, setter, deleter


def nested_descriptors(knob: str, knob_key_alias: str):
    """Generate descriptors for accessing a nested configuration knob.

    Args:
        knob: Nested configuration knob to generate descriptors for.
        knob_key_alias: Alias of knob key that needs to pbe defined in
            register before accessing nested configuration knobs.
    """

    def getter(self):
        try:
            primary_key = list(self._register.keys())[0]
            return self._register[primary_key].get(knob, None)
        except IndexError:
            raise KeyError(f"{knob_key_alias} must be defined before {knob} can be accessed.")

    def setter(self, value):
        try:
            primary_key = list(self._register.keys())[0]
            self._register[primary_key][knob] = value
        except IndexError:
            raise KeyError(f"{knob_key_alias} must be defined before {knob} can be accessed.")

    def deleter(self):
        try:
            primary_key = list(self._register.keys())[0]
            del self._register[primary_key][knob]
        except IndexError:
            raise KeyError(f"{knob_key_alias} must be defined before {knob} can be accessed.")
        except KeyError:
            pass

    return getter, setter, deleter


# Callbacks are used during parsing and marshalling for performing
# extra processing on specific configuration knob values. They contain callables
# that accept a single argument. Makes it easy to convert Python objects to Slurm
# configuration values and vice versa.
class Callback(NamedTuple):
    """Object for invoking callables on Slurm configuration knobs during parsing/marshalling.

    Possible callables:
        parse: Invoked when value is being parsed in from configuration file.
        marshal: Invoked when value is being marshalled into configuration file.
    """

    parse: Optional[Callable[[Any], Any]] = None
    marshal: Optional[Callable[[Any], Any]] = None


# Common parsing/marshalling callbacks for Slurm configuration values.
# Arrays are denoted using comma/colon separators. Maps are denoted as
# key1=value,key2=value,bool. Booleans are mapped by the inclusion of
# the keyword in maps. So key1=value,key2 would equate to:
#
# {
#   "key1": "value",
#   "key2": True,
# }
@functools.singledispatch
def _slurm_dict(v):
    raise TypeError(f"Expected str or dict, not {type(v)}")


@_slurm_dict.register
def _(v: str):
    """Convert Slurm dictionary to Python dictionary."""
    result = {}
    for val in v.split(","):
        if "=" in val:
            sub_opt, sub_val = val.split("=", 1)
            result.update({sub_opt: sub_val})
        else:
            result.update({val: True})

    return result


@_slurm_dict.register
def _(v: dict):
    """Convert Python dictionary to Slurm dictionary."""
    result = []
    for sub_opt, sub_val in v.items():
        if not isinstance(sub_val, bool):
            result.append(f"{sub_opt}={sub_val}")
        elif sub_val:
            result.append(sub_opt)

    return ",".join(result)


CommaSeparatorCallback = Callback(lambda v: v.split(","), lambda v: ",".join(v))
ColonSeparatorCallback = Callback(lambda v: v.split(":"), lambda v: ":".join(v))
SlurmDictCallback = Callback(_slurm_dict, _slurm_dict)
ReasonCallback = Callback(None, lambda v: f'"{v}"')


# All Slurm data models should inherit from this abstract parent class.
# The class provides method definitions for common operations and
# requires models to specify callbacks so that models can be treated
# generically when parsing and marshalling rather than having an infinite if-else tree.
class BaseModel(ABC):
    """Abstract base class for Slurm-related data models."""

    def __init__(self, **kwargs):
        self._register = kwargs

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            f"({', '.join(f'{k}={v}' for k, v in self._register.items())})"
        )

    @property
    @abstractmethod
    def _primary_key(self) -> Optional[str]:
        """Primary key for data model.

        A primary key is required for data models that have a unique identifier
        to preserve the integrity of the Slurm configuration syntax. For example,
        for compute nodes, the primary key would be the node name `NodeName`. Node
        name can be used nicely for identifying nodes in maps, but it is difficult to
        carry along the NodeName key inside the internal register of the class.

        _primary_key is used to track what the Slurm configuration key should be for
        unique identifiers. Without this "protected" attribute, we would likely need
        to write a custom parser for each data model. The generic model marshaller can
        detect this attribute and marshal the model accordingly.
        """
        pass

    @property
    @abstractmethod
    def _callbacks(self) -> MappingProxyType:
        """Store callbacks.

        This map will be queried during parsing and marshalling to determine if
        a configuration value needs any further processing. Each model class will
        need to define the callbacks specific to its configuration knobs. Every model
        class should declare whether it has callbacks or not.

        Callbacks should be MappingProxyType (read-only dict) to prevent any accidental
        mutation of callbacks used during parsing and marshalling.
        """
        pass

    def dict(self) -> Dict:
        """Get model in dictionary form.

        Returns a deep copy of model's internal register. The deep copy is needed
        because assigned variables all point to the same dictionary in memory. Without the
        deep copy, operations performed on the returned dictionary could cause unintended
        mutations in the internal register.
        """
        return copy.deepcopy(self._register)

    def json(self) -> str:
        """Get model as JSON object."""
        return json.dumps(self._register)



# Copyright 2024 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Edit slurm.conf files."""

__all__ = ["dump", "dumps", "load", "loads", "edit"]

import functools
import os
from collections import deque
from contextlib import contextmanager
from datetime import datetime
from typing import Union

from slurmutils.models import DownNodes, FrontendNode, Node, NodeSet, Partition, SlurmConfig

from ._editor import (
    clean,
    dump_base,
    dumps_base,
    header,
    load_base,
    loads_base,
    marshal_model,
    parse_model,
    parse_repeating_config,
)


def _marshaller(config: SlurmConfig) -> str:
    """Marshal Python object into slurm.conf configuration file.

    Args:
        config: `SlurmConfig` object to convert to slurm.conf configuration file.
    """
    marshalled = [header(f"`slurm.conf` file generated at {datetime.now()} by slurmutils.")]

    if config.include:
        marshalled.append(header("Included configuration files"))
        marshalled.extend([f"Include {i}\n" for i in config.include] + ["\n"])
    if config.slurmctld_host:
        marshalled.extend([f"SlurmctldHost={host}\n" for host in config.slurmctld_host] + ["\n"])

    # Marshal the SlurmConfig object into Slurm configuration format.
    # Ignore pockets containing child models as they will be marshalled inline.
    marshalled.extend(
        marshal_model(
            config,
            ignore={
                "Includes",
                "SlurmctldHost",
                "nodes",
                "frontend_nodes",
                "down_nodes",
                "node_sets",
                "partitions",
            },
        )
        + ["\n"]
    )

    if len(config.nodes) != 0:
        marshalled.extend(
            [header("Node configurations")]
            + [marshal_model(node, inline=True) for node in config.nodes]
            + ["\n"]
        )

    if len(config.frontend_nodes) != 0:
        marshalled.extend(
            [header("Frontend node configurations")]
            + [marshal_model(frontend, inline=True) for frontend in config.frontend_nodes]
            + ["\n"]
        )

    if len(config.down_nodes) != 0:
        marshalled.extend(
            [header("Down node configurations")]
            + [marshal_model(down_node, inline=True) for down_node in config.down_nodes]
            + ["\n"]
        )

    if len(config.node_sets) != 0:
        marshalled.extend(
            [header("Node set configurations")]
            + [marshal_model(node_set, inline=True) for node_set in config.node_sets]
            + ["\n"]
        )

    if len(config.partitions) != 0:
        marshalled.extend(
            [header("Partition configurations")]
            + [marshal_model(part, inline=True) for part in config.partitions]
        )

    return "".join(marshalled)


def _parser(config: str) -> SlurmConfig:
    """Parse slurm.conf configuration file into Python object.

    Args:
        config: Content of slurm.conf configuration file.
    """
    slurm_conf = {}
    nodes = {}
    frontend_nodes = {}
    down_nodes = []
    node_sets = {}
    partitions = {}

    config = clean(deque(config.splitlines()))
    while config:
        line = config.popleft()
        # slurm.conf `Include` is the only configuration knob whose
        # separator is whitespace rather than `=`.
        if line.startswith("Include"):
            option, value = line.split(maxsplit=1)
            parse_repeating_config(option, value, pocket=slurm_conf)

        # `SlurmctldHost` is the same as `Include` where it can
        # be specified on multiple lines.
        elif line.startswith("SlurmctldHost"):
            option, value = line.split("=", 1)
            parse_repeating_config(option, value, pocket=slurm_conf)

        # Check if option maps to slurm.conf data model. If so, invoke parsing
        # rules for that specific data model and enter its parsed information
        # into the appropriate pocket.
        elif line.startswith("NodeName"):
            parse_model(line, pocket=nodes, model=Node)
        elif line.startswith("FrontendNode"):
            parse_model(line, pocket=frontend_nodes, model=FrontendNode)
        elif line.startswith("DownNodes"):
            parse_model(line, pocket=down_nodes, model=DownNodes)
        elif line.startswith("NodeSet"):
            parse_model(line, pocket=node_sets, model=NodeSet)
        elif line.startswith("PartitionName"):
            parse_model(line, pocket=partitions, model=Partition)
        else:
            parse_model(line, pocket=slurm_conf, model=SlurmConfig)

    return SlurmConfig(
        **slurm_conf,
        nodes=nodes,
        frontend_nodes=frontend_nodes,
        down_nodes=down_nodes,
        node_sets=node_sets,
        partitions=partitions,
    )


dump = functools.partial(dump_base, marshaller=_marshaller)
dumps = functools.partial(dumps_base, marshaller=_marshaller)
load = functools.partial(load_base, parser=_parser)
loads = functools.partial(loads_base, parser=_parser)


@contextmanager
def edit(file: Union[str, os.PathLike]) -> SlurmConfig:
    """Edit a slurm.conf file.

    Args:
        file: Path to slurm.conf file to edit. If slurm.conf does
            not exist at the specified file path, it will be created.
    """
    if not os.path.exists(file):
        # Create an empty SlurmConfig that can be populated.
        config = SlurmConfig()
    else:
        config = load(file=file)

    yield config
    dump(content=config, file=file)



# Copyright 2024 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Edit slurmdbd.conf files."""

__all__ = ["dump", "dumps", "load", "loads", "edit"]

import functools
import os
from collections import deque
from contextlib import contextmanager
from datetime import datetime
from typing import Union

from slurmutils.models import SlurmdbdConfig

from ._editor import (
    clean,
    dump_base,
    dumps_base,
    header,
    load_base,
    loads_base,
    marshal_model,
    parse_model,
)


def _marshaller(config: SlurmdbdConfig) -> str:
    """Marshal Python object into slurmdbd.conf configuration file.

    Args:
        config: `SlurmdbdConfig` object to convert to slurmdbd.conf configuration file.
    """
    marshalled = [header(f"`slurmdbd.conf` file generated at {datetime.now()} by slurmutils.")]
    marshalled.extend(marshal_model(config))

    return "".join(marshalled)


def _parser(config: str) -> SlurmdbdConfig:
    """Parse slurmdbd.conf configuration file into Python object.

    Args:
        config: Content of slurmdbd.conf configuration file.
    """
    slurmdbd_conf = {}

    config = clean(deque(config.splitlines()))
    while config:
        line = config.popleft()
        parse_model(line, pocket=slurmdbd_conf, model=SlurmdbdConfig)

    return SlurmdbdConfig(**slurmdbd_conf)


dump = functools.partial(dump_base, marshaller=_marshaller)
dumps = functools.partial(dumps_base, marshaller=_marshaller)
load = functools.partial(load_base, parser=_parser)
loads = functools.partial(loads_base, parser=_parser)


@contextmanager
def edit(file: Union[str, os.PathLike]) -> SlurmdbdConfig:
    """Edit a slurmdbd.conf file.

    Args:
        file: Path to slurmdbd.conf file to edit. If slurmdbd.conf does
            not exist at the specified file path, it will be created.
    """
    if not os.path.exists(file):
        # Create an empty SlurmConfig that can be populated.
        config = SlurmdbdConfig()
    else:
        config = load(file=file)

    yield config
    dump(content=config, file=file)



# Copyright 2024 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Base methods for Slurm workload manager configuration file editors."""

import logging
import re
import shlex
from collections import deque
from os import PathLike
from pathlib import Path
from typing import Deque, Dict, List, Optional, Set, Union

from slurmutils.exceptions import EditorError

_logger = logging.getLogger(__name__)


def dump_base(content, file: Union[str, PathLike], marshaller):
    """Dump configuration into file using provided marshalling function.

    Do not use this function directly.
    """
    if (loc := Path(file)).exists():
        _logger.warning("Overwriting contents of %s file located at %s.", loc.name, loc)

    _logger.debug("Marshalling configuration into %s file located at %s.", loc.name, loc)
    return loc.write_text(marshaller(content), encoding="ascii")


def dumps_base(content, marshaller) -> str:
    """Dump configuration into Python string using provided marshalling function.

    Do not use this function directly.
    """
    return marshaller(content)


def load_base(file: Union[str, PathLike], parser):
    """Load configuration from file using provided parsing function.

    Do not use this function directly.
    """
    if (file := Path(file)).exists():
        _logger.debug("Parsing contents of %s located at %s.", file.name, file)
        config = file.read_text(encoding="ascii")
        return parser(config)
    else:
        msg = "Unable to locate file"
        _logger.error(msg + " %s.", file)
        raise FileNotFoundError(msg + f" {file}")


def loads_base(content: str, parser):
    """Load configuration from Python String using provided parsing function.

    Do not use this function directly.
    """
    return parser(content)


# Helper functions for parsing and marshalling Slurm configuration data.

_loose_pascal_filter = re.compile(r"(.)([A-Z][a-z]+)")
_snakecase_convert = re.compile(r"([a-z0-9])([A-Z])")


def _pascal2snake(v: str) -> str:
    """Convert string in loose PascalCase to snakecase.

    This private method takes in Slurm configuration knob keys and converts
    them to snakecase. The returned snakecase representation is used to
    dynamically access Slurm data model attributes and retrieve callbacks.
    """
    # The precompiled regex filters do a good job of converting Slurm's
    # loose PascalCase to snakecase, however, there are still some tokens
    # that slip through such as `CPUs`. This filter identifies those problematic
    # tokens and converts them into tokens that can be easily processed by the
    # compiled regex expressions.
    if "CPUs" in v:
        v = v.replace("CPUs", "Cpus")
    holder = _loose_pascal_filter.sub(r"\1_\2", v)
    return _snakecase_convert.sub(r"\1_\2", holder).lower()


def clean(config: Deque[str]) -> Deque[str]:
    """Clean loaded configuration file before parsing.

    Cleaning tasks include:
      1. Stripping away comments (#) in configuration. Slurm does not
         support octothorpes in strings; only for inline and standalone
         comments. **Do not use** octothorpes in Slurm configuration knob
         values as Slurm will treat anything proceeding an octothorpe as a comment.
      2. Strip away any extra whitespace at the end of each line.

    Args:
        config: Loaded configuration file. Split by newlines.
    """
    processed = deque()
    while config:
        line = config.popleft()
        if line.startswith("#"):
            # Skip comment lines as they're not necessary for configuration.
            continue
        elif "#" in line:
            # Slice off inline comment and strip away extra whitespace.
            processed.append(line[: line.index("#")].strip())
        else:
            processed.append(line.strip())

    return processed


def header(msg: str) -> str:
    """Generate header for marshalled configuration file.

    Args:
        msg: Message to put into header.
    """
    return "#\n" + "".join(f"# {line}\n" for line in msg.splitlines()) + "#\n"


def parse_repeating_config(__key, __value, pocket: Dict) -> None:
    """Parse `slurm.conf` configuration knobs with keys that can repeat.

    Args:
        __key: Configuration knob key that can repeat.
        __value: Value of the current configuration knob.
        pocket: Dictionary to add parsed configuration knob to.
    """
    if __key not in pocket:
        pocket[__key] = [__value]
    else:
        pocket[__key].append(__value)


def parse_model(line: str, pocket: Union[Dict, List], model) -> None:
    """Parse configuration knobs based on Slurm models.

    Model callbacks will be used for invoking special
    parsing if required for the configuration value in line.

    Args:
         line: Configuration line to parse.
         pocket: Dictionary to add parsed configuration knob to.
         model: Slurm data model to use for invoking callbacks and validating knob keys.
    """
    holder = {}
    for token in shlex.split(line):  # Use `shlex.split(...)` to preserve quotation blocks.
        # Word in front of the first `=` denotes the parent configuration knob key.
        option, value = token.split("=", maxsplit=1)
        if hasattr(model, attr := _pascal2snake(option)):
            if attr in model._callbacks and (callback := model._callbacks[attr].parse) is not None:
                holder.update({option: callback(value)})
            else:
                holder.update({option: value})
        else:
            raise EditorError(
                f"{option} is not a valid configuration option for {model.__name__}."
            )

    # Use temporary model object to update pocket with a Python dictionary
    # in the format that we want the dictionary to be.
    if isinstance(pocket, list):
        pocket.append(model(**holder).dict())
    else:
        pocket.update(model(**holder).dict())


def marshal_model(
    model, ignore: Optional[Set] = None, inline: bool = False
) -> Union[List[str], str]:
    """Marshal a Slurm model back into its Slurm configuration syntax.

    Args:
        model: Slurm model object to marshal into Slurm configuration syntax.
        ignore: Set of keys to ignore on model object when marshalling. Useful for models that
            have child models under certain keys that are directly handled. Default is None.
        inline: If True, marshal object into single line rather than multiline. Default is False.
    """
    marshalled = []
    if ignore is None:
        # Create an empty set if not ignores are specified. Prevents us from needing to
        # rely on a mutable default in the function signature.
        ignore = set()

    if primary_key := model._primary_key:
        attr = _pascal2snake(primary_key)
        primary_value = getattr(model, attr)
        data = {primary_key: primary_value, **model.dict()[primary_value]}
    else:
        data = model.dict()

    for option, value in data.items():
        if option not in ignore:
            if hasattr(model, attr := _pascal2snake(option)):
                if (
                    attr in model._callbacks
                    and (callback := model._callbacks[attr].marshal) is not None
                ):
                    value = callback(value)

                marshalled.append(f"{option}={value}")
            else:
                raise EditorError(
                    f"{option} is not a valid configuration option for {model.__class__.__name__}."
                )
        else:
            _logger.debug("Ignoring option %s. Option is present in ignore set %s", option, ignore)

    if inline:
        # Whitespace is the separator in Slurm configuration syntax.
        marshalled = " ".join(marshalled) + "\n"
    else:
        # Append newline character so that each configuration is on its own line.
        marshalled = [line + "\n" for line in marshalled]

    return marshalled
